#!/usr/bin/perl
# vdrtranscode server Version 0.1

# great thanks to this superb howto :
# http://trac.handbrake.fr/wiki/CLIGuide

use strict;
use warnings;
use Proc::Daemon;
use File::Find ;
use File::Copy ;
use File::Basename ;
use Getopt::Long ;
use Cwd;
use Logfile::Rotate;


my $use = "vdrtranscode_server.pl

\$ vdrtranscode_server [--daemon][--log][--verbose][--help]

" ;


## GetOpt::Long 
my $daemon_flag ;
my $verbose_flag ;
my $log_flag ;


GetOptions( 	"daemon" 	=>	\$daemon_flag ,
			"log" 	=>		\$log_flag ,
			"verbose" => 	\$verbose_flag,
			"help"	=> 	sub { print "$use" ; exit 0 ;}
) ;
my $config = &parse_config("/etc/vdrtranscode.conf") ;

# change user to vdr running User  if called as root
if ( $> == 0 ) {
  my $uid = getpwnam($config->{vdr_user});
  $> = $uid ;
  &message("change effective User ID to $config->{vdr_user} : $uid") ;
}


my $Indir = "$config->{Indir}" ;
my $Outdir = "$config->{Outdir}" ;

if ( $daemon_flag ) {
  undef $verbose_flag ;
  Proc::Daemon::Init;
}
else { &message("running in foreground...") };
chdir($Outdir) ; 

# leave our process ID for init
open MYPID , (">/tmp/vdrtranscode_server.pid") ;
print MYPID $$ ;
close MYPID ; 

# find binarys 
my $hb_bin = `which HandBrakeCLI` ; # eg. /usr/bin/HandBrakeCLI
chomp $hb_bin ;

# remove old logfile on server start
#if ( $log_flag ) { 
#    open LOG , ">./vdrtranscode_server.log" ;
#    close LOG ;
#}

my $continue = 1;
#$SIG{TERM} = sub { $continue = 0 };
$SIG{TERM} = \&quit ; # exit clean on Signal kill PID
$SIG{INT} = \&quit ; # exit clean on Signal Str+C


# declarations
my @VideoList ;
my $workfile ;

while ($continue) {

# log rotation , to prevent large Logfiles
if ( -f "./vdrtranscode_server.log" and -s "./vdrtranscode_server.log" > 1024000 ) { # log file larger than 1024 Kbyte 
  message("rotate Logfiles...") ;
  my $gzip_bin =`which gzip`; chomp $gzip_bin ;
  my $logdatei = new Logfile::Rotate(
                    File  => "./vdrtranscode_server.log",
                    Count => 5,
                    Gzip  => "$gzip_bin",
                   );
  $logdatei->rotate();
  undef $logdatei;
}

# File find [cut- in /video
# reset on every new main loop
@VideoList =() ;
$workfile ="" ;

find ( \&funcfind_index , "$Indir" )  ; # alle /.*\/\[cut\].*\/.*rec\/index/  Dateien finden
# if yes
  if ( $#VideoList >= 0 ) {
     &message("found $VideoList[0]") ;
    #mark [work
    &rename_vdr_file( $VideoList[0] , "cut" , "work") ;
    ( $workfile = $VideoList[0] ) =~ s/\[cut/\[work/ ; # looks like /video/24/[work-HDTV]08.24-15:00_-_16:00_Uhr/2010-06-08.03.36.5-0.rec/

    if ( -f "${workfile}00001.ts" ) { 
      ${workfile} = "${workfile}00001.ts" ;
    }
    # breaks loop if no 00001.ts exists
    else {
      &rename_vdr_file(${workfile} , "work" , "oops" ) ;
      &message("oops no 00001.ts found in $VideoList[0]") ;
      next ;
    }
    # analyse File
    &message("analysiere...") ;
    my $workfile_dosh = dosh($workfile) ;

    # pre analysis to get fps for calculating start point
    my $fps ="" ;
    open PRE_ANALYZER , "nice -n $config->{nice_level} $hb_bin -i $workfile_dosh -o /dev/null -t 0  2>&1 |" ;
    	while ( my $Zeile = <PRE_ANALYZER> ) {
	    if ( $Zeile =~/size: / ) {
	      ( undef , $fps ) = $Zeile  =~/size:\s(\d+x\d+),.*,\s(\d+)\.\d+\sfps/ ;
	      &message("\$fps $fps") ;
	    }
	}
    close PRE_ANALYZER ;

    # check marks and resolve start and end point
    # looks like : 
    # 0:05:36.25
    # 0:48:21.04
    my $param_start = "" ;
    my $param_stop = "" ;
    my $Durframes = 0 ;
    my ( $mark_dir ) = ${workfile} =~ /(.*)00001.ts/ ;
    if ( -f "${mark_dir}marks" ) {
	open MARKS , "<${mark_dir}marks" ; my @marks = <MARKS> ; close MARKS ;
	 if ( $#marks == 1 ) {
	    my $z = 0 ;
	    foreach ( @marks ) {
	      $z++ ;
	      my ( $Hour , $Minute , $Second , $Frame ) = $_ =~/(\d+):(\d+):(\d+).(\d+)/ ;
	      $param_start =  ( ($Hour * 60 *60 *$fps) + ($Minute * 60 *$fps) + ($Second *$fps)+ $Frame)  if $z == 1 ;
	      $param_stop =  ( ($Hour * 60 *60 *$fps) + ($Minute * 60 *$fps) + ($Second *$fps)+ $Frame) if $z == 2 ;
	    }
	  # stop frame is counted from start frame, not from Filebegin
	  $param_stop = $param_stop - $param_start ;
	  $Durframes = $param_stop ;
	  $param_stop = "--stop-at frame:$param_stop" ;
	  $param_start = "--start-at frame:$param_start" ;
	}
    }
    &message("\$param_start $param_start\n\$param_stop $param_stop") ;

    open ANALYZER , "nice -n $config->{nice_level} $hb_bin -i $workfile_dosh -o /dev/null $param_start -t 0  2>&1 |" ;

    # declarations 
    my $follow_audiotracks = 0 ;
    my $WxH = "" ;
    my ( $hours , $minutes , $seconds ) = "" ;
    my @Atracks = () ;
	while ( my $Zeile = <ANALYZER> ) {
	  # find  informations
	  # + duration: 00:55:41
	  if ( $Zeile =~/duration: / ) { 
	    ( $hours , $minutes , $seconds ) = $Zeile  =~/(\d+):(\d+):(\d+)/ ;
	  }
	  # + size: 1920x1080, pixel aspect: 1/1, display aspect: 1.78, 25.000 fps   
	  if ( $Zeile =~/size: / ) { 
	    ( $WxH , undef ) = $Zeile  =~/size:\s(\d+x\d+),.*,\s(\d+)\.\d+\sfps/ ;
	    &message("$WxH") ;
	    # calculate frames all over
	    if ( $Durframes == 0 ) { # no markers used 
		$Durframes = (( $hours * 60 * 60 )+ ( $minutes *60 ) + $seconds ) * $fps  ;
		&message("\$Durframes neu $Durframes") ;
	    }
	  }
	  # + audio tracks:
	  if ( $Zeile =~/audio tracks:/ ) { $follow_audiotracks = 1 ; }
	  if ( $Zeile =~/1,|2,|3,|4,/ and $follow_audiotracks == 1 ) { 
#	    print "$Zeile" ;
	    my ( $nr , $lang , $codec  ) = $Zeile  =~/\+\s(\d+),\s+(\w+)\s+\((\w+)\)/ ;
	      my $kbps = "" ;
	      if ( $Zeile =~ /AC3/ ) { ( $kbps )  = $Zeile  =~/,\s+(\d+)bps/ ; $kbps = $kbps / 1000 ; } # for Ac3 files get bitrate in kbps 
#	    print "$nr , $lang , $codec , $kbps\n" ;
	    @{$Atracks[$nr]}  = ( $lang , $codec , $kbps ) ; # structure Array[Tracknumber]->[field 0 : language] , [field 1 : codec] , [field 2: kbs]
	  # + 1, Deutsch (AC3) (2.0 ch) (iso639-2: deu), 48000Hz, 384000bps
	  # + 2, English (AC3) (2.0 ch) (iso639-2: eng), 48000Hz, 384000bps
	  }
	}
    close ANALYZER ;

# get Informations for processing from File Flag
# [mp4|m4v|mkv] [DD|noDD|HD-HD|HD-smallHD] [UVHQ|VHQ|HQ|MQ|LQ] [first|all]

  my ( $container , $dd_hd_sd , $quali , $atracks ) = $workfile =~ /\[work-(mp4|m4v|mkv)\|(DD|noDD|HD-HD|HD-smallHD)\|(UVHQ|VHQ|HQ|MQ|LQ)\|(first|all)\]/ ;
  &message("$container , $dd_hd_sd , $quali , $atracks") ;

# build cmd line 
# audiopart
# -a 1,1,2 -A "Main Audio","Downmixed Audio","Director's Commentary"-E ac3,aac,aac -B auto,160,128 -R auto,auto,44100 -6 auto,dpl2,stereo 
# which audiotracks ?
my $param_a ="" ; # Orig. Audio Tracks to use
my $param_A ="" ; # Audio description
my $param_E ="" ; #  Audio Encoder
my $param_B ="" ; # Audio Bitrates
my $param_D ="" ;# normalize Audio 
my $nr_of_mp2 = 0 ;
my $nr_of_mp2_used = 0 ;
my @arr_of_track_contain_mp2 =() ;
my $nr_of_ac3 = 0 ;
my $nr_of_ac3_used = 0 ;
my $ac3_bitrate = 0 ;
my @arr_of_track_contain_ac3 =() ;
    foreach my $i ( 1..$#Atracks ) {
	#structure Array[Tracknumber]->[field 0 : language] , [field 1 : codec] , [field 2: kbs]
	if ( $Atracks[$i][1] =~/mp2/ ) { $nr_of_mp2++ ; push @arr_of_track_contain_mp2 , $i }
	if ( $Atracks[$i][1] =~/AC3/ ) { $nr_of_ac3++ ; push @arr_of_track_contain_ac3 , $i }
	&message("Atracks : $Atracks[$i][0], $Atracks[$i][1], $Atracks[$i][2]") ;
    }
# mp2
if ($atracks eq "first" and $nr_of_mp2 >=1 ) { 
      $param_a = "$arr_of_track_contain_mp2[0]," ; 
      $param_A = "\"$Atracks[$arr_of_track_contain_mp2[0]][0]\"" ;
      $param_E = "faac," ;
      $param_B = "$config->{AAC_Bitrate}," ;
      $param_D = "$config->{DRC}," ;
      $nr_of_mp2_used++ ;
}
if ($atracks eq "all" and $nr_of_mp2 >=1 ) {
      foreach  ( @arr_of_track_contain_mp2 ) {
	$param_a .= "${_}," ;
	$param_A .="\"$Atracks[${_}][0]\"," ;
	$param_E .= "faac," ;
	$param_B .= "$config->{AAC_Bitrate}," ;
	$param_D .= "$config->{DRC}," ;
	$nr_of_mp2_used++ ;
     }
}
# ac3
if ($dd_hd_sd =~ /DD|HD-HD|HD-smallHD/ and $nr_of_ac3 >=1) {
      foreach  ( @arr_of_track_contain_ac3 ) {
	$param_a .= "${_}," ;
	$param_A .="\"$Atracks[${_}][0]\"," ;
	$param_E .= "copy," ;
	$param_B .= "auto," ;
	$param_D .= "1.0," ;
	$nr_of_ac3_used++ ;
	$ac3_bitrate = "$Atracks[${_}][2]" ;
	
      }
}
foreach ( $param_a, $param_A , $param_E , $param_B , $param_D ) { $_ =~s/,$// ; } # remove last komma

&message("\$param_a -a $param_a\n\$param_A -A $param_A\n\$param_E -E $param_E\n\$param_B -B $param_B\n\$param_D -D $param_D") ;

# test to prevent jerky playing on 25 fps sources
my $set_fps="" ;
if ( $fps == 25 ) { $set_fps ="-r $fps"} 

my $x264_opts ="ref=2:mixed-refs:bframes=2:b-pyramid=1:weightb=1:analyse=all:8x8dct=1:subme=7:me=umh:merange=24:trellis=1:no-fast-pskip=1:no-dct-decimate=1:direct=auto" ;

# Picture Size
# $dd_hd_sd holds one of this -> DD|noDD|HD-HD|HD-smallHD
my $param_X = 720 ; # max. Dimension Width ( 720 SD ( default )  , 1280 smallHD , 1920 HD )
$param_X = 1920 if ( $dd_hd_sd eq "HD-HD" ) ;
$param_X = 1280 if ( $dd_hd_sd eq "HD-smallHD" ) ;

# anamorphic encoding 
my $param_anamorph = "" ;
if ( $config->{anamorph_encoding} == 1 ) { $param_anamorph = "--loose-anamorphic" } # enable anamorph_encoding

# LQ for Webencoding -> sets maximum width of picture to 480 , disables anamorph encoding, sets AAC Rate to 96
if ( $quali eq "LQ" ) {
  $param_X = 480 ;
  $param_anamorph = "" ;
# TODO current all Audiotracks are passed from general Settings above , override here 
#  $ac3_bitrate = 96 ; ##TODO
  ${container} = "m4v" if ( $nr_of_ac3 > 0 ) ; # change container from mp4 to m4v , if  ac3 avaible
}

  my $outfile = "${Outdir}vdrtrancode_tmp.${container}" ;
# recalculate Videobitrate to match round Mbyte Sizes ( cosmetic programming )
# $frames 
# $fps 
# $aac_nr 
# $aac_bitrate 
# $ac3_nr 
# $ac3_bitrate 
# $wish_bitrate 
my $recalc_video_bitrate = &recalculate_video_bitrate( $Durframes , $fps , $nr_of_mp2_used , $config->{AAC_Bitrate}, $nr_of_ac3_used , $ac3_bitrate , $config->{$quali}) ;

## strucure of proccesing line
## HandBrakeCLI -i /video/Wir_sind_Kaiser_-_Best_of/2010-10-26.21.55.15-0.rec/00001.ts  -o ./test3.mp4 -e x264 -O -b 500 -2 -T -x ref=2:mixed-refs:bframes=2:b-pyramid=1:
## weightb=1:analyse=all:8x8dct=1:subme=7:me=umh:merange=24:trellis=1:no-fast-pskip=1:no-dct-decimate=1:direct=auto -5 -B 128  --stop-at frame:3000 --strict-anamorphic
 
 my $start_to_log = 0 ; 
 #open HBLOG , ">>./hb_log" ;

# overwrite for debug
#$param_stop = "--stop-at frame:3000" ;
 &message("JOBSTART --- $workfile") ;
 open HB , "nice -n $config->{nice_level} $hb_bin -i $workfile_dosh -O $set_fps -b $recalc_video_bitrate -2 -T  -e x264 -x $x264_opts -5 -a $param_a -A $param_A -E $param_E -B $param_B -D $param_D $param_anamorph --modulus 8 -X $param_X -o $outfile $param_start $param_stop 2>&1 1>/dev/null |" ;
    while ( my $Zeile = <HB> ) {
     &message("° $Zeile") ;
#      if ( $Zeile =~/job configuration:/ ) { $start_to_log = 1 ; &message("JOBSTART --- $workfile") } # start logging
#      if ( $Zeile =~/reader: first SCR/ ) { $start_to_log = 0 ; } # stop logging
#      if ( $start_to_log == 1 ) { &message("$Zeile") }  

    }
  my $time_now = localtime ;
  &message("$time_now JOBEND --- $workfile") ;
#  close HBLOG ;
  close HB ;  

  # mark [del
  ( $workfile = $workfile ) =~s/00001.ts// ;
   &rename_vdr_file( $workfile , "work" , "del") ;

#rename outfile
# $workfile looks like #  "/video/Der_Terminator/[work-m4v|HD-smallHD|VHQ|all]Science-Fiction/2010-06-08.03.36.5-0.rec/" 
my $copy_workfile = $workfile ;
$copy_workfile =~s/\/\d{4}-\d{2}-\d{2}.*rec\/// ;
$copy_workfile =~s/\[work.*(first|all)\]// ;
$copy_workfile =~s/$Indir// ;
$copy_workfile =~s/\//-/ ;
$copy_workfile .= ".${container}" ;

rename ("$outfile" , "${Outdir}$copy_workfile") ;
 
#  exit 0 ;
  }
# if no "[cut-" File found
 else {
  # wait 60 seconds
&message("waiting") ;
  sleep 60 ;
  }

# functions inside loop
#################################################################
sub dosh {
my $in = $_[0] ;
#$in =~ s/\s+/\ /g ; ## doppelte oder mehrere Leerzeichen auf eins reduzieren
$in =~ s/(\(\d+)\/(\d+\))/${1}\/${2}/g ;       ## das (1/5) problem
$in =~ s/([\s \( \) \$ \& \§ \" \! \? \[  \] \' \,\@ \| \> \<])/\\$1/g; # viele absonderliche Sonderzeichen für die shell qouten
$in =~ s/\\\\/\\/g ; # wenn ausdruck vorher schon geqotet war, die doppelten backslashes wieder auf einen kompenisieren
return $in ;
}
#################################################################
  sub funcfind_index {
  return unless ( $File::Find::name =~ /.*\/\[cut-.*\].*\/.*rec\/index$/ ) ;
  $File::Find::name=~s/index// ;
  push ( @VideoList , $File::Find::name ) ;
  }
#################################################################
 sub rename_vdr_file {
  # /video/24/[cut-HDTV]08.24-15:00_-_16:00_Uhr/2010-06-08.03.36.5-0.rec
  ( my $filename = $_[0] ) =~ s/\d{4}-\d{2}-.*\.rec// ;
  my $from = $_[1] ;
  my $to = $_[2] ;
  my $dir = dirname($filename) ; 
  my $file = basename($filename) ;
    ( my $file_neu = $file )=~ s/\[$from/\[$to/ ;
    &message("rename : ${dir}/${file} , ${dir}/${file_neu}") ;
  rename ( "${dir}/${file}" , "${dir}/${file_neu}" ) ;
  open REFRESH , ">$Indir/.update" ;
  close REFRESH ;
  }
#################################################################
 sub recalculate_video_bitrate {
  my $frames = shift ; 
  my $fps = shift ;
  my $aac_nr = shift ;
  my $aac_bitrate = shift ;
  my $ac3_nr = shift ;
  my $ac3_bitrate = shift ;
  my $wish_bitrate = shift ;

  # Calculate Size of AAC Files
  my $AudioKbyte ;
  if ( $aac_nr > 0 ) {
    $AudioKbyte = sprintf ( "%.8f" , ( $aac_bitrate * $frames / $fps / 8 ) ) ; # ohne Overhead
    if ( $aac_nr >= 1 ) { $AudioKbyte = $AudioKbyte *  $aac_nr }
  }
  else { $AudioKbyte = 0 }
#  print "\$AudioKbyte $AudioKbyte\n" ;
   
  # Calculate Size of Ac3 Files
  my	$ac3_kbyte_sec ;
  my $ac3_kbyteSize ;
  if ( $ac3_nr > 0 ) {
    $ac3_kbyte_sec = $ac3_bitrate / 8 ;
    $ac3_kbyteSize = sprintf ( "%.8f" , ( $ac3_kbyte_sec  * $frames  / $fps )) ;
    if ( $ac3_nr >= 1 ) { $ac3_kbyteSize = $ac3_kbyteSize * $ac3_nr }
  }
  else  { $ac3_kbyteSize = 0 }
#  print "\$ac3_kbyteSize $ac3_kbyteSize\n" ;


  # Memory count 
  my $minutes = $frames * $fps * 60 ;
  my $video_kbyte_sec = $wish_bitrate / 8 ; 
  my $video_kbyte_size = ( $video_kbyte_sec * $frames ) / $fps ; 
  my $probe_memory_ammount_kbyte = $video_kbyte_size + $ac3_kbyteSize + $AudioKbyte ;
  my $probe_memory_ammount_Mbyte = sprintf "%i" , $probe_memory_ammount_kbyte / 1024 ;
  &message("\$probe_memory_ammount_Mbyte $probe_memory_ammount_Mbyte") ;

  my $round_memory_ammount_Mbyte ;
  # rounding
  if ( $probe_memory_ammount_Mbyte < 100 ) {
	  $round_memory_ammount_Mbyte = ( sprintf "%.0f" , ( $probe_memory_ammount_Mbyte / 5 )) * 5  ;
#	  if ( $round_memory_ammount_Mbyte < $probe_memory_ammount_Mbyte ){ $round_memory_ammount_Mbyte+=5}
  }
  elsif ( $probe_memory_ammount_Mbyte < 600 ) {
	  $round_memory_ammount_Mbyte = ( sprintf "%.0f" , ( $probe_memory_ammount_Mbyte / 25 )) * 25 ;
#	  if ( $round_memory_ammount_Mbyte < $probe_memory_ammount_Mbyte ){ $round_memory_ammount_Mbyte+=25}
  }
  elsif ( $probe_memory_ammount_Mbyte >= 600 ) {
	  $round_memory_ammount_Mbyte = ( sprintf "%.0f" , ( $probe_memory_ammount_Mbyte / 50 )) * 50  ;
#	  if ( $round_memory_ammount_Mbyte < $probe_memory_ammount_Mbyte ){ $round_memory_ammount_Mbyte+=50}
  }
  &message("\$round_memory_ammount_Mbyte $round_memory_ammount_Mbyte") ;
  my $round_memory_ammount_kbyte = $round_memory_ammount_Mbyte * 1024 ;
  $round_memory_ammount_kbyte = $round_memory_ammount_kbyte * 1.018 ; # empiric factor add 
  my $round_memory_ammount_video_kbit = ( $round_memory_ammount_kbyte - $ac3_kbyteSize - $AudioKbyte ) * 8 ;
  my $round_memory_ammount_video_kbit_sec = sprintf "%.0f" , $round_memory_ammount_video_kbit / ( $frames / $fps ) ;
  &message("\$round_memory_ammount_video_kbit_sec $round_memory_ammount_video_kbit_sec") ;
  return $round_memory_ammount_video_kbit_sec ; 
}
#################################################################
# end while} 
}

# global functions
#################################################################
sub quit {
  message("*quit...") ;
  unlink ("/tmp/vdrtranscode_server.pid") if ( -f "/tmp/vdrtranscode_server.pid" ) ;
  die  ; 
}
#################################################################
# thanks to http://www.patshaping.de/hilfen_ta/codeschnipsel/perl-configparser.htm
sub parse_config($)
{
 my $file = shift;
 local *CF;

 open(CF,'<'.$file) or die "Open $file: $!";
 read(CF, my $data, -s $file);
 close(CF);

 my @lines  = split(/\015\012|\012|\015/,$data);
 my $config = {};
 my $count  = 0;

 foreach my $line(@lines)
 {
  $count++;

  next if($line =~ /^\s*#/);
  next if($line !~ /^\s*\S+\s*=.*$/);

  my ($key,$value) = split(/=/,$line,2);

  # Remove whitespaces at the beginning and at the end

  $key   =~ s/^\s+//g;
  $key   =~ s/\s+$//g;
  $value =~ s/^\s+//g;
  $value =~ s/\s+$//g;

#  die "Configuration option '$key' defined twice in line $count of configuration file '$file'" if($config->{$key});

  $config->{$key} = $value;
 }

 return $config;
}
#################################################################
sub message {
  # consider if message goes to STDOUT , Logfile or /dev/zero
   my $message = shift ;
   chomp $message ;
  if ( $verbose_flag and not $daemon_flag ) {
    print "$message\n" ;
  }
  if ( $log_flag ) {
    open LOG , ">>./vdrtranscode_server.log" ;
    unless ( $message =~/waiting/ ) { print LOG "$message\n" ; } # dont flood log with "waiting"
    close LOG ;
   }
}